from __future__ import print_function
import netlink.capi as nl
import netlink.genl.capi as genl
import nl80211
import sys
import traceback


class test_class:
    def __init__(self):
        self.done = 1


def freq_to_ch(freq):
    if freq == 2484:
        return 14

    if freq < 2484:
        return (freq - 2407) / 5

    # FIXME: dot11ChannelStartingFactor (802.11-2007 17.3.8.3.2)
    if freq < 45000:
        return freq / 5 - 1000

    if freq >= 58320 and freq <= 64800:
        return (freq - 56160) / 2160

    return 0


def handle_freq(attr, pol):
    e, fattr = nl.py_nla_parse_nested(nl80211.NL80211_FREQUENCY_ATTR_MAX, attr, pol)
    if nl80211.NL80211_FREQUENCY_ATTR_FREQ in fattr:
        freq = nl.nla_get_u32(fattr[nl80211.NL80211_FREQUENCY_ATTR_FREQ])
        sys.stdout.write("\t\tfreq %d MHz [%d]" % (freq, freq_to_ch(freq)))
    if nl80211.NL80211_FREQUENCY_ATTR_MAX_TX_POWER in fattr and not (
        nl80211.NL80211_FREQUENCY_ATTR_DISABLED in fattr
    ):
        sys.stdout.write(
            " (%.1f dBm)"
            % (
                0.01
                * nl.nla_get_u32(fattr[nl80211.NL80211_FREQUENCY_ATTR_MAX_TX_POWER])
            )
        )
    if nl80211.NL80211_FREQUENCY_ATTR_DISABLED in fattr:
        sys.stdout.write(" (disabled)")
    sys.stdout.write("\n")


def handle_band(attr, fpol):
    e, battr = nl.py_nla_parse_nested(nl80211.NL80211_BAND_ATTR_MAX, attr, None)
    print("\tband %d:" % nl.nla_type(attr))
    if nl80211.NL80211_BAND_ATTR_FREQS in battr:
        for fattr in nl.nla_get_nested(battr[nl80211.NL80211_BAND_ATTR_FREQS]):
            handle_freq(fattr, fpol)


def cipher_name(suite):
    suite_val = "%02x%02x%02x%02x" % tuple(reversed(suite))
    if suite_val == "000fac01":
        return "WEP40 (00-0f-ac:1)"
    elif suite_val == "000fac05":
        return "WEP104 (00-0f-ac:5)"
    elif suite_val == "000fac02":
        return "TKIP (00-0f-ac:2)"
    elif suite_val == "000fac04":
        return "CCMP (00-0f-ac:4)"
    elif suite_val == "000fac06":
        return "CMAC (00-0f-ac:6)"
    elif suite_val == "000fac08":
        return "GCMP (00-0f-ac:8)"
    elif suite_val == "00147201":
        return "WPI-SMS4 (00-14-72:1)"
    else:
        return suite_val


def msg_handler(m, a):
    try:
        e, attr = genl.py_genlmsg_parse(
            nl.nlmsg_hdr(m), 0, nl80211.NL80211_ATTR_MAX, None
        )
        if nl80211.NL80211_ATTR_WIPHY_NAME in attr:
            print("wiphy %s" % nl.nla_get_string(attr[nl80211.NL80211_ATTR_WIPHY_NAME]))
        if nl80211.NL80211_ATTR_WIPHY_BANDS in attr:
            fpol = nl.nla_policy_array(nl80211.NL80211_FREQUENCY_ATTR_MAX + 1)
            fpol[nl80211.NL80211_FREQUENCY_ATTR_FREQ].type = nl.NLA_U32
            fpol[nl80211.NL80211_FREQUENCY_ATTR_DISABLED].type = nl.NLA_FLAG
            fpol[nl80211.NL80211_FREQUENCY_ATTR_PASSIVE_SCAN].type = nl.NLA_FLAG
            fpol[nl80211.NL80211_FREQUENCY_ATTR_NO_IBSS].type = nl.NLA_FLAG
            fpol[nl80211.NL80211_FREQUENCY_ATTR_RADAR].type = nl.NLA_FLAG
            fpol[nl80211.NL80211_FREQUENCY_ATTR_MAX_TX_POWER].type = nl.NLA_U32

            nattrs = nl.nla_get_nested(attr[nl80211.NL80211_ATTR_WIPHY_BANDS])
            for nattr in nattrs:
                handle_band(nattr, fpol)
        if nl80211.NL80211_ATTR_CIPHER_SUITES in attr:
            ciphers = nl.nla_data(attr[nl80211.NL80211_ATTR_CIPHER_SUITES])
            num = len(ciphers) / 4
            if num > 0:
                print("\tSupported Ciphers:")
                for i in range(0, num, 4):
                    print("\t\t* %s" % cipher_name(ciphers[i : i + 4]))
        if nl80211.NL80211_ATTR_SUPPORTED_IFTYPES in attr:
            print("\tSupported interface modes:")
            ifattr = nl.nla_get_nested(attr[nl80211.NL80211_ATTR_SUPPORTED_IFTYPES])
            for nl_mode in ifattr:
                print("\t\t* %s" % nl80211.nl80211_iftype2str[nl.nla_type(nl_mode)])
        if nl80211.NL80211_ATTR_SOFTWARE_IFTYPES in attr:
            print("\tsoftware interface modes (can always be added):")
            ifattr = nl.nla_get_nested(attr[nl80211.NL80211_ATTR_SOFTWARE_IFTYPES])
            for nl_mode in ifattr:
                print("\t\t* %s" % nl80211.nl80211_iftype2str[nl.nla_type(nl_mode)])
        return nl.NL_SKIP
    except Exception:
        (t, v, tb) = sys.exc_info()
        print(v.message)
        traceback.print_tb(tb)


def error_handler(err, a):
    a.done = err.error
    return nl.NL_STOP


def finish_handler(m, a):
    return nl.NL_SKIP


def ack_handler(m, a):
    a.done = 0
    return nl.NL_STOP


try:
    cbd = test_class()
    tx_cb = nl.nl_cb_alloc(nl.NL_CB_DEFAULT)
    rx_cb = nl.nl_cb_clone(tx_cb)
    s = nl.nl_socket_alloc_cb(tx_cb)
    nl.py_nl_cb_err(rx_cb, nl.NL_CB_CUSTOM, error_handler, cbd)
    nl.py_nl_cb_set(rx_cb, nl.NL_CB_FINISH, nl.NL_CB_CUSTOM, finish_handler, cbd)
    nl.py_nl_cb_set(rx_cb, nl.NL_CB_ACK, nl.NL_CB_CUSTOM, ack_handler, cbd)
    nl.py_nl_cb_set(rx_cb, nl.NL_CB_VALID, nl.NL_CB_CUSTOM, msg_handler, cbd)

    genl.genl_connect(s)
    family = genl.genl_ctrl_resolve(s, "nl80211")
    m = nl.nlmsg_alloc()
    genl.genlmsg_put(m, 0, 0, family, 0, 0, nl80211.NL80211_CMD_GET_WIPHY, 0)
    nl.nla_put_u32(m, nl80211.NL80211_ATTR_WIPHY, 7)

    err = nl.nl_send_auto_complete(s, m)
    if err < 0:
        nl.nlmsg_free(m)

    while cbd.done > 0 and not err < 0:
        err = nl.nl_recvmsgs(s, rx_cb)
except Exception:
    (t, v, tb) = sys.exc_info()
    print(v.message)
    traceback.print_tb(tb)



#
# Copyright (c) 2011 Thomas Graf <tgraf@suug.ch>
#

"""Module providing access to network links

This module provides an interface to view configured network links,
modify them and to add and delete virtual network links.

The following is a basic example:
    import netlink.core as netlink
    import netlink.route.link as link

    sock = netlink.Socket()
    sock.connect(netlink.NETLINK_ROUTE)

    cache = link.LinkCache()	# create new empty link cache
    cache.refill(sock)		# fill cache with all configured links
    eth0 = cache['eth0']		# lookup link "eth0"
    print eth0			# print basic configuration

The module contains the following public classes:

  - Link -- Represents a network link. Instances can be created directly
        via the constructor (empty link objects) or via the refill()
        method of a LinkCache.
  - LinkCache -- Derived from netlink.Cache, holds any number of
         network links (Link instances). Main purpose is to keep
         a local list of all network links configured in the
         kernel.

The following public functions exist:
  - get_from_kernel(socket, name)

"""

from __future__ import absolute_import

__version__ = "0.1"
__all__ = [
    "LinkCache",
    "Link",
]

import socket
from .. import core as netlink
from .. import capi as core_capi
from . import capi as capi
from .links import inet as inet
from .. import util as util

# Link statistics definitions
RX_PACKETS = 0
TX_PACKETS = 1
RX_BYTES = 2
TX_BYTES = 3
RX_ERRORS = 4
TX_ERRORS = 5
RX_DROPPED = 6
TX_DROPPED = 7
RX_COMPRESSED = 8
TX_COMPRESSED = 9
RX_FIFO_ERR = 10
TX_FIFO_ERR = 11
RX_LEN_ERR = 12
RX_OVER_ERR = 13
RX_CRC_ERR = 14
RX_FRAME_ERR = 15
RX_MISSED_ERR = 16
TX_ABORT_ERR = 17
TX_CARRIER_ERR = 18
TX_HBEAT_ERR = 19
TX_WIN_ERR = 20
COLLISIONS = 21
MULTICAST = 22
IP6_INPKTS = 23
IP6_INHDRERRORS = 24
IP6_INTOOBIGERRORS = 25
IP6_INNOROUTES = 26
IP6_INADDRERRORS = 27
IP6_INUNKNOWNPROTOS = 28
IP6_INTRUNCATEDPKTS = 29
IP6_INDISCARDS = 30
IP6_INDELIVERS = 31
IP6_OUTFORWDATAGRAMS = 32
IP6_OUTPKTS = 33
IP6_OUTDISCARDS = 34
IP6_OUTNOROUTES = 35
IP6_REASMTIMEOUT = 36
IP6_REASMREQDS = 37
IP6_REASMOKS = 38
IP6_REASMFAILS = 39
IP6_FRAGOKS = 40
IP6_FRAGFAILS = 41
IP6_FRAGCREATES = 42
IP6_INMCASTPKTS = 43
IP6_OUTMCASTPKTS = 44
IP6_INBCASTPKTS = 45
IP6_OUTBCASTPKTS = 46
IP6_INOCTETS = 47
IP6_OUTOCTETS = 48
IP6_INMCASTOCTETS = 49
IP6_OUTMCASTOCTETS = 50
IP6_INBCASTOCTETS = 51
IP6_OUTBCASTOCTETS = 52
ICMP6_INMSGS = 53
ICMP6_INERRORS = 54
ICMP6_OUTMSGS = 55
ICMP6_OUTERRORS = 56


class LinkCache(netlink.Cache):
    """Cache of network links"""

    def __init__(self, family=socket.AF_UNSPEC, cache=None):
        if not cache:
            cache = self._alloc_cache_name("route/link")

        self._info_module = None
        self._protocol = netlink.NETLINK_ROUTE
        self._nl_cache = cache
        self._set_arg1(family)

    def __getitem__(self, key):
        if type(key) is int:
            link = capi.rtnl_link_get(self._nl_cache, key)
        else:
            link = capi.rtnl_link_get_by_name(self._nl_cache, key)

        if link is None:
            raise KeyError()
        else:
            return Link.from_capi(link)

    @staticmethod
    def _new_object(obj):
        return Link(obj)

    def _new_cache(self, cache):
        return LinkCache(family=self.arg1, cache=cache)


class Link(netlink.Object):
    """Network link"""

    def __init__(self, obj=None):
        netlink.Object.__init__(self, "route/link", "link", obj)
        self._rtnl_link = self._obj2type(self._nl_object)

        if self.type:
            self._module_lookup("netlink.route.links." + self.type)

        self.inet = inet.InetLink(self)
        self.af = {"inet": self.inet}

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_value, tb):
        if exc_type is None:
            self.change()
        else:
            return False

    @classmethod
    def from_capi(cls, obj):
        return cls(capi.link2obj(obj))

    @staticmethod
    def _obj2type(obj):
        return capi.obj2link(obj)

    def __cmp__(self, other):
        return self.ifindex - other.ifindex

    @staticmethod
    def _new_instance(obj):
        if not obj:
            raise ValueError()

        return Link(obj)

    @property
    @netlink.nlattr(type=int, immutable=True, fmt=util.num)
    def ifindex(self):
        """interface index"""
        return capi.rtnl_link_get_ifindex(self._rtnl_link)

    @ifindex.setter
    def ifindex(self, value):
        capi.rtnl_link_set_ifindex(self._rtnl_link, int(value))

        # ifindex is immutable but we assume that if _orig does not
        # have an ifindex specified, it was meant to be given here
        if capi.rtnl_link_get_ifindex(self._orig) == 0:
            capi.rtnl_link_set_ifindex(self._orig, int(value))

    @property
    @netlink.nlattr(type=str, fmt=util.bold)
    def name(self):
        """Name of link"""
        return capi.rtnl_link_get_name(self._rtnl_link)

    @name.setter
    def name(self, value):
        capi.rtnl_link_set_name(self._rtnl_link, value)

        # name is the secondary identifier, if _orig does not have
        # the name specified yet, assume it was meant to be specified
        # here. ifindex will always take priority, therefore if ifindex
        # is specified as well, this will be ignored automatically.
        if capi.rtnl_link_get_name(self._orig) is None:
            capi.rtnl_link_set_name(self._orig, value)

    @property
    @netlink.nlattr(type=str, fmt=util.string)
    def flags(self):
        """Flags
        Setting this property will *Not* reset flags to value you supply in
        Examples:
        link.flags = '+xxx' # add xxx flag
        link.flags = 'xxx'  # exactly the same
        link.flags = '-xxx' # remove xxx flag
        link.flags = [ '+xxx', '-yyy' ] # list operation
        """
        flags = capi.rtnl_link_get_flags(self._rtnl_link)
        return capi.rtnl_link_flags2str(flags, 256)[0].split(",")

    def _set_flag(self, flag):
        if flag.startswith("-"):
            i = capi.rtnl_link_str2flags(flag[1:])
            capi.rtnl_link_unset_flags(self._rtnl_link, i)
        elif flag.startswith("+"):
            i = capi.rtnl_link_str2flags(flag[1:])
            capi.rtnl_link_set_flags(self._rtnl_link, i)
        else:
            i = capi.rtnl_link_str2flags(flag)
            capi.rtnl_link_set_flags(self._rtnl_link, i)

    @flags.setter
    def flags(self, value):
        if not (type(value) is str):
            for flag in value:
                self._set_flag(flag)
        else:
            self._set_flag(value)

    @property
    @netlink.nlattr(type=int, fmt=util.num)
    def mtu(self):
        """Maximum Transmission Unit"""
        return capi.rtnl_link_get_mtu(self._rtnl_link)

    @mtu.setter
    def mtu(self, value):
        capi.rtnl_link_set_mtu(self._rtnl_link, int(value))

    @property
    @netlink.nlattr(type=int, immutable=True, fmt=util.num)
    def family(self):
        """Address family"""
        return capi.rtnl_link_get_family(self._rtnl_link)

    @family.setter
    def family(self, value):
        capi.rtnl_link_set_family(self._rtnl_link, value)

    @property
    @netlink.nlattr(type=str, fmt=util.addr)
    def address(self):
        """Hardware address (MAC address)"""
        a = capi.rtnl_link_get_addr(self._rtnl_link)
        return netlink.AbstractAddress(a)

    @address.setter
    def address(self, value):
        capi.rtnl_link_set_addr(self._rtnl_link, value._addr)

    @property
    @netlink.nlattr(type=str, fmt=util.addr)
    def broadcast(self):
        """Hardware broadcast address"""
        a = capi.rtnl_link_get_broadcast(self._rtnl_link)
        return netlink.AbstractAddress(a)

    @broadcast.setter
    def broadcast(self, value):
        capi.rtnl_link_set_broadcast(self._rtnl_link, value._addr)

    @property
    @netlink.nlattr(type=str, immutable=True, fmt=util.string)
    def qdisc(self):
        """Name of qdisc (cannot be changed)"""
        return capi.rtnl_link_get_qdisc(self._rtnl_link)

    @qdisc.setter
    def qdisc(self, value):
        capi.rtnl_link_set_qdisc(self._rtnl_link, value)

    @property
    @netlink.nlattr(type=int, fmt=util.num)
    def txqlen(self):
        """Length of transmit queue"""
        return capi.rtnl_link_get_txqlen(self._rtnl_link)

    @txqlen.setter
    def txqlen(self, value):
        capi.rtnl_link_set_txqlen(self._rtnl_link, int(value))

    @property
    @netlink.nlattr(type=str, immutable=True, fmt=util.string)
    def arptype(self):
        """Type of link (cannot be changed)"""
        type_ = capi.rtnl_link_get_arptype(self._rtnl_link)
        return core_capi.nl_llproto2str(type_, 64)[0]

    @arptype.setter
    def arptype(self, value):
        i = core_capi.nl_str2llproto(value)
        capi.rtnl_link_set_arptype(self._rtnl_link, i)

    @property
    @netlink.nlattr(type=str, immutable=True, fmt=util.string, title="state")
    def operstate(self):
        """Operational status"""
        operstate = capi.rtnl_link_get_operstate(self._rtnl_link)
        return capi.rtnl_link_operstate2str(operstate, 32)[0]

    @operstate.setter
    def operstate(self, value):
        i = capi.rtnl_link_str2operstate(value)
        capi.rtnl_link_set_operstate(self._rtnl_link, i)

    @property
    @netlink.nlattr(type=str, immutable=True, fmt=util.string)
    def mode(self):
        """Link mode"""
        mode = capi.rtnl_link_get_linkmode(self._rtnl_link)
        return capi.rtnl_link_mode2str(mode, 32)[0]

    @mode.setter
    def mode(self, value):
        i = capi.rtnl_link_str2mode(value)
        capi.rtnl_link_set_linkmode(self._rtnl_link, i)

    @property
    @netlink.nlattr(type=str, fmt=util.string)
    def alias(self):
        """Interface alias (SNMP)"""
        return capi.rtnl_link_get_ifalias(self._rtnl_link)

    @alias.setter
    def alias(self, value):
        capi.rtnl_link_set_ifalias(self._rtnl_link, value)

    @property
    @netlink.nlattr(type=str, fmt=util.string)
    def type(self):
        """Link type"""
        return capi.rtnl_link_get_type(self._rtnl_link)

    @type.setter
    def type(self, value):
        if capi.rtnl_link_set_type(self._rtnl_link, value) < 0:
            raise NameError("unknown info type")

        self._module_lookup("netlink.route.links." + value)

    def get_stat(self, stat):
        """Retrieve statistical information"""
        if type(stat) is str:
            stat = capi.rtnl_link_str2stat(stat)
            if stat < 0:
                raise NameError("unknown name of statistic")

        return capi.rtnl_link_get_stat(self._rtnl_link, stat)

    def enslave(self, slave, sock=None):
        if not sock:
            sock = netlink.lookup_socket(netlink.NETLINK_ROUTE)

        return capi.rtnl_link_enslave(sock._sock, self._rtnl_link, slave._rtnl_link)

    def release(self, slave, sock=None):
        if not sock:
            sock = netlink.lookup_socket(netlink.NETLINK_ROUTE)

        return capi.rtnl_link_release(sock._sock, self._rtnl_link, slave._rtnl_link)

    def add(self, sock=None, flags=None):
        if not sock:
            sock = netlink.lookup_socket(netlink.NETLINK_ROUTE)

        if not flags:
            flags = netlink.NLM_F_CREATE

        ret = capi.rtnl_link_add(sock._sock, self._rtnl_link, flags)
        if ret < 0:
            raise netlink.KernelError(ret)

    def change(self, sock=None, flags=0):
        """Commit changes made to the link object"""
        if sock is None:
            sock = netlink.lookup_socket(netlink.NETLINK_ROUTE)

        if not self._orig:
            raise netlink.NetlinkError("Original link not available")
        ret = capi.rtnl_link_change(sock._sock, self._orig, self._rtnl_link, flags)
        if ret < 0:
            raise netlink.KernelError(ret)

    def delete(self, sock=None):
        """Attempt to delete this link in the kernel"""
        if sock is None:
            sock = netlink.lookup_socket(netlink.NETLINK_ROUTE)

        ret = capi.rtnl_link_delete(sock._sock, self._rtnl_link)
        if ret < 0:
            raise netlink.KernelError(ret)

    ###################################################################
    # private properties
    #
    # Used for formatting output. USE AT OWN RISK
    @property
    def _state(self):
        if "up" in self.flags:
            buf = util.good("up")
            if "lowerup" not in self.flags:
                buf += " " + util.bad("no-carrier")
        else:
            buf = util.bad("down")
        return buf

    @property
    def _brief(self):
        return self._module_brief() + self._foreach_af("brief")

    @property
    def _flags(self):
        ignore = [
            "up",
            "running",
            "lowerup",
        ]
        return ",".join([flag for flag in self.flags if flag not in ignore])

    def _foreach_af(self, name, args=None):
        buf = ""
        for af in self.af:
            try:
                func = getattr(self.af[af], name)
                s = str(func(args))
                if len(s) > 0:
                    buf += " " + s
            except AttributeError:
                pass
        return buf

    def format(self, details=False, stats=False, indent=""):
        """Return link as formatted text"""
        fmt = util.MyFormatter(self, indent)

        buf = fmt.format(
            "{a|ifindex} {a|name} {a|arptype} {a|address} "
            "{a|_state} <{a|_flags}> {a|_brief}"
        )

        if details:
            buf += fmt.nl("\t{t|mtu} {t|txqlen} {t|weight} " "{t|qdisc} {t|operstate}")
            buf += fmt.nl("\t{t|broadcast} {t|alias}")

            buf += self._foreach_af("details", fmt)

        if stats:
            lst = [
                ["Packets", RX_PACKETS, TX_PACKETS],
                ["Bytes", RX_BYTES, TX_BYTES],
                ["Errors", RX_ERRORS, TX_ERRORS],
                ["Dropped", RX_DROPPED, TX_DROPPED],
                ["Compressed", RX_COMPRESSED, TX_COMPRESSED],
                ["FIFO Errors", RX_FIFO_ERR, TX_FIFO_ERR],
                ["Length Errors", RX_LEN_ERR, None],
                ["Over Errors", RX_OVER_ERR, None],
                ["CRC Errors", RX_CRC_ERR, None],
                ["Frame Errors", RX_FRAME_ERR, None],
                ["Missed Errors", RX_MISSED_ERR, None],
                ["Abort Errors", None, TX_ABORT_ERR],
                ["Carrier Errors", None, TX_CARRIER_ERR],
                ["Heartbeat Errors", None, TX_HBEAT_ERR],
                ["Window Errors", None, TX_WIN_ERR],
                ["Collisions", None, COLLISIONS],
                ["Multicast", None, MULTICAST],
                ["", None, None],
                ["Ipv6:", None, None],
                ["Packets", IP6_INPKTS, IP6_OUTPKTS],
                ["Bytes", IP6_INOCTETS, IP6_OUTOCTETS],
                ["Discards", IP6_INDISCARDS, IP6_OUTDISCARDS],
                ["Multicast Packets", IP6_INMCASTPKTS, IP6_OUTMCASTPKTS],
                ["Multicast Bytes", IP6_INMCASTOCTETS, IP6_OUTMCASTOCTETS],
                ["Broadcast Packets", IP6_INBCASTPKTS, IP6_OUTBCASTPKTS],
                ["Broadcast Bytes", IP6_INBCASTOCTETS, IP6_OUTBCASTOCTETS],
                ["Delivers", IP6_INDELIVERS, None],
                ["Forwarded", None, IP6_OUTFORWDATAGRAMS],
                ["No Routes", IP6_INNOROUTES, IP6_OUTNOROUTES],
                ["Header Errors", IP6_INHDRERRORS, None],
                ["Too Big Errors", IP6_INTOOBIGERRORS, None],
                ["Address Errors", IP6_INADDRERRORS, None],
                ["Unknown Protocol", IP6_INUNKNOWNPROTOS, None],
                ["Truncated Packets", IP6_INTRUNCATEDPKTS, None],
                ["Reasm Timeouts", IP6_REASMTIMEOUT, None],
                ["Reasm Requests", IP6_REASMREQDS, None],
                ["Reasm Failures", IP6_REASMFAILS, None],
                ["Reasm OK", IP6_REASMOKS, None],
                ["Frag Created", None, IP6_FRAGCREATES],
                ["Frag Failures", None, IP6_FRAGFAILS],
                ["Frag OK", None, IP6_FRAGOKS],
                ["", None, None],
                ["ICMPv6:", None, None],
                ["Messages", ICMP6_INMSGS, ICMP6_OUTMSGS],
                ["Errors", ICMP6_INERRORS, ICMP6_OUTERRORS],
            ]

            buf += "\n\t%s%s%s%s\n" % (
                33 * " ",
                util.title("RX"),
                15 * " ",
                util.title("TX"),
            )

            for row in lst:
                row[0] = util.kw(row[0])
                row[1] = self.get_stat(row[1]) if row[1] else ""
                row[2] = self.get_stat(row[2]) if row[2] else ""
                buf += "\t{0[0]:27} {0[1]:>16} {0[2]:>16}\n".format(row)

            buf += self._foreach_af("stats")

        return buf


def get(name, sock=None):
    """Lookup Link object directly from kernel"""
    if not name:
        raise ValueError()

    if not sock:
        sock = netlink.lookup_socket(netlink.NETLINK_ROUTE)

    link = capi.get_from_kernel(sock._sock, 0, name)
    if not link:
        return None

    return Link.from_capi(link)


_link_cache = LinkCache()


def resolve(name):
    _link_cache.refill()
    return _link_cache[name]



#
# Copyright (c) 2011 Thomas Graf <tgraf@suug.ch>
#

"""VLAN network link

"""

from __future__ import absolute_import


from ... import core as netlink
from .. import capi as capi


class VLANLink(object):
    def __init__(self, link):
        self._link = link

    @property
    @netlink.nlattr(type=int)
    def id(self):
        """vlan identifier"""
        return capi.rtnl_link_vlan_get_id(self._link)

    @id.setter
    def id(self, value):
        capi.rtnl_link_vlan_set_id(self._link, int(value))

    @property
    @netlink.nlattr(type=str)
    def flags(self):
        """VLAN flags
        Setting this property will *Not* reset flags to value you supply in
        Examples:
        link.flags = '+xxx' # add xxx flag
        link.flags = 'xxx'  # exactly the same
        link.flags = '-xxx' # remove xxx flag
        link.flags = [ '+xxx', '-yyy' ] # list operation
        """
        flags = capi.rtnl_link_vlan_get_flags(self._link)
        return capi.rtnl_link_vlan_flags2str(flags, 256)[0].split(",")

    def _set_flag(self, flag):
        if flag.startswith("-"):
            i = capi.rtnl_link_vlan_str2flags(flag[1:])
            capi.rtnl_link_vlan_unset_flags(self._link, i)
        elif flag.startswith("+"):
            i = capi.rtnl_link_vlan_str2flags(flag[1:])
            capi.rtnl_link_vlan_set_flags(self._link, i)
        else:
            i = capi.rtnl_link_vlan_str2flags(flag)
            capi.rtnl_link_vlan_set_flags(self._link, i)

    @flags.setter
    def flags(self, value):
        if type(value) is list:
            for flag in value:
                self._set_flag(flag)
        else:
            self._set_flag(value)

    ###################################################################
    # TODO:
    #   - ingress map
    #   - egress map

    def brief(self):
        return "vlan-id {0}".format(self.id)


def init(link):
    link.vlan = VLANLink(link._rtnl_link)
    return link.vlan



#
# Utilities
#
# Copyright (c) 2011 Thomas Graf <tgraf@suug.ch>
#

"""utility module for netlink

"""

from __future__ import absolute_import

from . import capi as capi
from string import Formatter

__version__ = "1.0"


# rename into colored_output
def _color(t, c):
    return "{esc}[{color}m{text}{esc}[0m".format(esc=b"\x1b".decode(), color=c, text=t)


def black(t):
    return _color(t, 30)


def red(t):
    return _color(t, 31)


def green(t):
    return _color(t, 32)


def yellow(t):
    return _color(t, 33)


def blue(t):
    return _color(t, 34)


def magenta(t):
    return _color(t, 35)


def cyan(t):
    return _color(t, 36)


def white(t):
    return _color(t, 37)


def bold(t):
    return _color(t, 1)


def kw(t):
    return yellow(t)


def num(t):
    return str(t)


def string(t):
    return t


def addr(t):
    return str(t)


def bad(t):
    return red(t)


def good(t):
    return green(t)


def title(t):
    return t


def boolean(t):
    return str(t)


def handle(t):
    return str(t)


class MyFormatter(Formatter):
    def __init__(self, obj, indent=""):
        self._obj = obj
        self._indent = indent

    def _nlattr(self, key):
        value = getattr(self._obj.__class__, key)
        if not isinstance(value, property):
            raise ValueError("Invalid formatting string {0}".format(key))

        d = getattr(value.fget, "formatinfo", {})

        # value = value.fget() is exactly the same
        value = getattr(self._obj, key)

        if "fmt" in d:
            value = d["fmt"](value)

        title_ = d.get("title", None)

        return title_, str(value)

    def get_value(self, key, args, kwds):
        # Let default get_value() handle ints
        if not isinstance(key, str):
            return Formatter.get_value(self, key, args, kwds)

        # HACK, we allow defining strings via fields to allow
        # conversions
        if key[:2] == "s|":
            return key[2:]

        if key[:2] == "t|":
            # title mode ("TITLE ATTR")
            include_title = True
        elif key[:2] == "a|":
            # plain attribute mode ("ATTR")
            include_title = False
        else:
            # No special field, have default get_value() get it
            return Formatter.get_value(self, key, args, kwds)

        key = key[2:]
        (title_, value) = self._nlattr(key)

        if include_title:
            if not title_:
                title_ = key  # fall back to key as title
            value = "{0} {1}".format(kw(title_), value)

        return value

    def convert_field(self, value, conversion):
        if conversion == "r":
            return repr(value)
        elif conversion == "s":
            return str(value)
        elif conversion == "k":
            return kw(value)
        elif conversion == "b":
            return bold(value)
        elif conversion is None:
            return value

        raise ValueError("Unknown converion specifier {0!s}".format(conversion))

    def nl(self, format_string=""):
        return "\n" + self._indent + self.format(format_string)


NL_BYTE_RATE = 0
NL_BIT_RATE = 1


class Rate(object):
    def __init__(self, rate, mode=NL_BYTE_RATE):
        self._rate = rate
        self._mode = mode

    def __str__(self):
        return capi.nl_rate2str(self._rate, self._mode, 32)[1]

    def __int__(self):
        return self._rate

    def __cmp__(self, other):
        return int(self) - int(other)


class Size(object):
    def __init__(self, size):
        self._size = size

    def __str__(self):
        return capi.nl_size2str(self._size, 32)[0]

    def __int__(self):
        return self._size

    def __cmp__(self, other):
        return int(self) - int(other)



from __future__ import print_function
import netlink.capi as nl
import netlink.genl.capi as genl
import nl80211
import sys
import traceback


class test_class:
    def __init__(self):
        self.done = 1


def msg_handler(m, a):
    try:
        e, attr = genl.py_genlmsg_parse(
            nl.nlmsg_hdr(m), 0, nl80211.NL80211_ATTR_MAX, None
        )
        if nl80211.NL80211_ATTR_WIPHY in attr:
            thiswiphy = nl.nla_get_u32(attr[nl80211.NL80211_ATTR_WIPHY])
            print("phy#%d" % thiswiphy)
        if nl80211.NL80211_ATTR_IFNAME in attr:
            print(
                "\tinterface %s" % nl.nla_get_string(attr[nl80211.NL80211_ATTR_IFNAME])
            )
        if nl80211.NL80211_ATTR_IFINDEX in attr:
            print("\tifindex %d" % nl.nla_get_u32(attr[nl80211.NL80211_ATTR_IFINDEX]))
        if nl80211.NL80211_ATTR_WDEV in attr:
            print("\twdev 0x%lx" % nl.nla_get_u64(attr[nl80211.NL80211_ATTR_WDEV]))
        if nl80211.NL80211_ATTR_MAC in attr:
            print(
                "\tmac %02x:%02x:%02x:%02x:%02x:%02x"
                % tuple(nl.nla_data(attr[nl80211.NL80211_ATTR_MAC]))
            )
        if nl80211.NL80211_ATTR_SSID in attr:
            print("\tssid ", nl.nla_data(attr[nl80211.NL80211_ATTR_SSID]))
        if nl80211.NL80211_ATTR_IFTYPE in attr:
            iftype = nl.nla_get_u32(attr[nl80211.NL80211_ATTR_IFTYPE])
            print("\ttype %s" % nl80211.nl80211_iftype2str[iftype])
        if nl80211.NL80211_ATTR_WIPHY_FREQ in attr:
            freq = nl.nla_get_u32(attr[nl80211.NL80211_ATTR_WIPHY_FREQ])

            sys.stdout.write("\tfreq %d MHz" % freq)

            if nl80211.NL80211_ATTR_CHANNEL_WIDTH in attr:
                chanw = nl.nla_get_u32(attr[nl80211.NL80211_ATTR_CHANNEL_WIDTH])
                sys.stdout.write(", width: %s" % nl80211.nl80211_chan_width2str[chanw])
                if nl80211.NL80211_ATTR_CENTER_FREQ1 in attr:
                    sys.stdout.write(
                        ", center1: %d MHz"
                        % nl.nla_get_u32(attr[nl80211.NL80211_ATTR_CENTER_FREQ1])
                    )
                if nl80211.NL80211_ATTR_CENTER_FREQ2 in attr:
                    sys.stdout.write(
                        ", center2: %d MHz"
                        % nl.nla_get_u32(attr[nl80211.NL80211_ATTR_CENTER_FREQ2])
                    )
            elif nl80211.NL80211_ATTR_WIPHY_CHANNEL_TYPE in attr:
                channel_type = nl.nla_get_u32(
                    attr[nl80211.NL80211_ATTR_WIPHY_CHANNEL_TYPE]
                )
                sys.stdout.write(" %s" % nl80211.nl80211_channel_type2str(channel_type))

            sys.stdout.write("\n")
        return nl.NL_SKIP
    except Exception:
        (t, v, tb) = sys.exc_info()
        print(v.message)
        traceback.print_tb(tb)


def error_handler(err, a):
    a.done = err.error
    return nl.NL_STOP


def finish_handler(m, a):
    return nl.NL_SKIP


def ack_handler(m, a):
    a.done = 0
    return nl.NL_STOP


try:
    cbd = test_class()
    tx_cb = nl.nl_cb_alloc(nl.NL_CB_DEFAULT)
    rx_cb = nl.nl_cb_clone(tx_cb)
    s = nl.nl_socket_alloc_cb(tx_cb)
    nl.py_nl_cb_err(rx_cb, nl.NL_CB_CUSTOM, error_handler, cbd)
    nl.py_nl_cb_set(rx_cb, nl.NL_CB_FINISH, nl.NL_CB_CUSTOM, finish_handler, cbd)
    nl.py_nl_cb_set(rx_cb, nl.NL_CB_ACK, nl.NL_CB_CUSTOM, ack_handler, cbd)
    nl.py_nl_cb_set(rx_cb, nl.NL_CB_VALID, nl.NL_CB_CUSTOM, msg_handler, cbd)

    genl.genl_connect(s)
    family = genl.genl_ctrl_resolve(s, "nl80211")
    m = nl.nlmsg_alloc()
    genl.genlmsg_put(m, 0, 0, family, 0, 0, nl80211.NL80211_CMD_GET_INTERFACE, 0)
    nl.nla_put_u32(m, nl80211.NL80211_ATTR_IFINDEX, nl.if_nametoindex("wlan0"))

    err = nl.nl_send_auto_complete(s, m)
    if err < 0:
        nl.nlmsg_free(m)

    while cbd.done > 0 and not err < 0:
        err = nl.nl_recvmsgs(s, rx_cb)

except Exception:
    (t, v, tb) = sys.exc_info()
    print(v.message)
    traceback.print_tb(tb)



from __future__ import print_function
import netlink.core as netlink
import netlink.route.capi as capi
import netlink.route.link as link

sock = netlink.lookup_socket(netlink.NETLINK_ROUTE)

cache = link.LinkCache()
cache.refill(sock)

testtap1 = cache["testtap1"]
print(testtap1)

lbr = link.Link()
lbr.type = "bridge"
lbr.name = "testbridge"
print(lbr)
lbr.add()

cache.refill(sock)
lbr = cache["testbridge"]
print(lbr)

lbr.enslave(testtap1)
cache.refill(sock)
testtap1 = cache["testtap1"]

print(capi.rtnl_link_is_bridge(lbr._rtnl_link))
print(capi.rtnl_link_get_master(testtap1._rtnl_link))



#
# Copyright (c) 2011 Thomas Graf <tgraf@suug.ch>
#

"""Module providing access to network addresses
"""

from __future__ import absolute_import


__version__ = "1.0"
__all__ = ["AddressCache", "Address"]

import datetime
from .. import core as netlink
from . import capi as capi
from . import link as Link
from .. import util as util


class AddressCache(netlink.Cache):
    """Cache containing network addresses"""

    def __init__(self, cache=None):
        if not cache:
            cache = self._alloc_cache_name("route/addr")

        self._protocol = netlink.NETLINK_ROUTE
        self._nl_cache = cache

    def __getitem__(self, key):
        # Using ifindex=0 here implies that the local address itself
        # is unique, otherwise the first occurence is returned.
        return self.lookup(0, key)

    def lookup(self, ifindex, local):
        if type(local) is str:
            local = netlink.AbstractAddress(local)

        addr = capi.rtnl_addr_get(self._nl_cache, ifindex, local._nl_addr)
        if addr is None:
            raise KeyError()

        return Address._from_capi(addr)

    @staticmethod
    def _new_object(obj):
        return Address(obj)

    @staticmethod
    def _new_cache(cache):
        return AddressCache(cache=cache)


class Address(netlink.Object):
    """Network address"""

    def __init__(self, obj=None):
        netlink.Object.__init__(self, "route/addr", "address", obj)
        self._rtnl_addr = self._obj2type(self._nl_object)

    @classmethod
    def _from_capi(cls, obj):
        return cls(capi.addr2obj(obj))

    @staticmethod
    def _obj2type(obj):
        return capi.obj2addr(obj)

    def __cmp__(self, other):
        # sort by:
        #    1. network link
        #    2. address family
        #    3. local address (including prefixlen)
        diff = self.ifindex - other.ifindex

        if diff == 0:
            diff = self.family - other.family
            if diff == 0:
                diff = capi.nl_addr_cmp(self.local, other.local)

        return diff

    @staticmethod
    def _new_instance(obj):
        return Address(obj)

    @property
    @netlink.nlattr(type=int, immutable=True, fmt=util.num)
    def ifindex(self):
        """interface index"""
        return capi.rtnl_addr_get_ifindex(self._rtnl_addr)

    @ifindex.setter
    def ifindex(self, value):
        link = Link.resolve(value)
        if not link:
            raise ValueError()

        self.link = link

    @property
    @netlink.nlattr(type=str, fmt=util.string)
    def link(self):
        link = capi.rtnl_addr_get_link(self._rtnl_addr)
        if not link:
            return None

        return Link.Link.from_capi(link)

    @link.setter
    def link(self, value):
        if type(value) is str:
            try:
                value = Link.resolve(value)
            except KeyError:
                raise ValueError()

        capi.rtnl_addr_set_link(self._rtnl_addr, value._rtnl_link)

        # ifindex is immutable but we assume that if _orig does not
        # have an ifindex specified, it was meant to be given here
        if capi.rtnl_addr_get_ifindex(self._orig) == 0:
            capi.rtnl_addr_set_ifindex(self._orig, value.ifindex)

    @property
    @netlink.nlattr(type=str, fmt=util.string)
    def label(self):
        """address label"""
        return capi.rtnl_addr_get_label(self._rtnl_addr)

    @label.setter
    def label(self, value):
        capi.rtnl_addr_set_label(self._rtnl_addr, value)

    @property
    @netlink.nlattr(type=str, fmt=util.string)
    def flags(self):
        """Flags

        Setting this property will *Not* reset flags to value you supply in

        Examples:
        addr.flags = '+xxx' # add xxx flag
        addr.flags = 'xxx'  # exactly the same
        addr.flags = '-xxx' # remove xxx flag
        addr.flags = [ '+xxx', '-yyy' ] # list operation
        """
        flags = capi.rtnl_addr_get_flags(self._rtnl_addr)
        return capi.rtnl_addr_flags2str(flags, 256)[0].split(",")

    def _set_flag(self, flag):
        if flag.startswith("-"):
            i = capi.rtnl_addr_str2flags(flag[1:])
            capi.rtnl_addr_unset_flags(self._rtnl_addr, i)
        elif flag.startswith("+"):
            i = capi.rtnl_addr_str2flags(flag[1:])
            capi.rtnl_addr_set_flags(self._rtnl_addr, i)
        else:
            i = capi.rtnl_addr_str2flags(flag)
            capi.rtnl_addr_set_flags(self._rtnl_addr, i)

    @flags.setter
    def flags(self, value):
        if type(value) is list:
            for flag in value:
                self._set_flag(flag)
        else:
            self._set_flag(value)

    @property
    @netlink.nlattr(type=int, immutable=True, fmt=util.num)
    def family(self):
        """Address family"""
        fam = capi.rtnl_addr_get_family(self._rtnl_addr)
        return netlink.AddressFamily(fam)

    @family.setter
    def family(self, value):
        if not isinstance(value, netlink.AddressFamily):
            value = netlink.AddressFamily(value)

        capi.rtnl_addr_set_family(self._rtnl_addr, int(value))

    @property
    @netlink.nlattr(type=int, fmt=util.num)
    def scope(self):
        """Address scope"""
        scope = capi.rtnl_addr_get_scope(self._rtnl_addr)
        return capi.rtnl_scope2str(scope, 32)[0]

    @scope.setter
    def scope(self, value):
        if type(value) is str:
            value = capi.rtnl_str2scope(value)
        capi.rtnl_addr_set_scope(self._rtnl_addr, value)

    @property
    @netlink.nlattr(type=str, immutable=True, fmt=util.addr)
    def local(self):
        """Local address"""
        a = capi.rtnl_addr_get_local(self._rtnl_addr)
        return netlink.AbstractAddress(a)

    @local.setter
    def local(self, value):
        a = netlink.AbstractAddress(value)
        capi.rtnl_addr_set_local(self._rtnl_addr, a._nl_addr)

        # local is immutable but we assume that if _orig does not
        # have a local address specified, it was meant to be given here
        if capi.rtnl_addr_get_local(self._orig) is None:
            capi.rtnl_addr_set_local(self._orig, a._nl_addr)

    @property
    @netlink.nlattr(type=str, fmt=util.addr)
    def peer(self):
        """Peer address"""
        a = capi.rtnl_addr_get_peer(self._rtnl_addr)
        return netlink.AbstractAddress(a)

    @peer.setter
    def peer(self, value):
        a = netlink.AbstractAddress(value)
        capi.rtnl_addr_set_peer(self._rtnl_addr, a._nl_addr)

    @property
    @netlink.nlattr(type=str, fmt=util.addr)
    def broadcast(self):
        """Broadcast address"""
        a = capi.rtnl_addr_get_broadcast(self._rtnl_addr)
        return netlink.AbstractAddress(a)

    @broadcast.setter
    def broadcast(self, value):
        a = netlink.AbstractAddress(value)
        capi.rtnl_addr_set_broadcast(self._rtnl_addr, a._nl_addr)

    @property
    @netlink.nlattr(type=str, fmt=util.addr)
    def multicast(self):
        """multicast address"""
        a = capi.rtnl_addr_get_multicast(self._rtnl_addr)
        return netlink.AbstractAddress(a)

    @multicast.setter
    def multicast(self, value):
        try:
            a = netlink.AbstractAddress(value)
        except ValueError as err:
            raise AttributeError("multicast", err)

        capi.rtnl_addr_set_multicast(self._rtnl_addr, a._nl_addr)

    @property
    @netlink.nlattr(type=str, fmt=util.addr)
    def anycast(self):
        """anycast address"""
        a = capi.rtnl_addr_get_anycast(self._rtnl_addr)
        return netlink.AbstractAddress(a)

    @anycast.setter
    def anycast(self, value):
        a = netlink.AbstractAddress(value)
        capi.rtnl_addr_set_anycast(self._rtnl_addr, a._nl_addr)

    @property
    @netlink.nlattr(type=int, immutable=True, fmt=util.num)
    def valid_lifetime(self):
        """Valid lifetime"""
        msecs = capi.rtnl_addr_get_valid_lifetime(self._rtnl_addr)
        if msecs == 0xFFFFFFFF:
            return None
        else:
            return datetime.timedelta(seconds=msecs)

    @valid_lifetime.setter
    def valid_lifetime(self, value):
        capi.rtnl_addr_set_valid_lifetime(self._rtnl_addr, int(value))

    @property
    @netlink.nlattr(type=int, immutable=True, fmt=util.num)
    def preferred_lifetime(self):
        """Preferred lifetime"""
        msecs = capi.rtnl_addr_get_preferred_lifetime(self._rtnl_addr)
        if msecs == 0xFFFFFFFF:
            return None
        else:
            return datetime.timedelta(seconds=msecs)

    @preferred_lifetime.setter
    def preferred_lifetime(self, value):
        capi.rtnl_addr_set_preferred_lifetime(self._rtnl_addr, int(value))

    @property
    @netlink.nlattr(type=int, immutable=True, fmt=util.num)
    def create_time(self):
        """Creation time"""
        hsec = capi.rtnl_addr_get_create_time(self._rtnl_addr)
        return datetime.timedelta(milliseconds=10 * hsec)

    @property
    @netlink.nlattr(type=int, immutable=True, fmt=util.num)
    def last_update(self):
        """Last update"""
        hsec = capi.rtnl_addr_get_last_update_time(self._rtnl_addr)
        return datetime.timedelta(milliseconds=10 * hsec)

    def add(self, socket=None, flags=None):
        if not socket:
            socket = netlink.lookup_socket(netlink.NETLINK_ROUTE)

        if not flags:
            flags = netlink.NLM_F_CREATE

        ret = capi.rtnl_addr_add(socket._sock, self._rtnl_addr, flags)
        if ret < 0:
            raise netlink.KernelError(ret)

    def delete(self, socket, flags=0):
        """Attempt to delete this address in the kernel"""
        ret = capi.rtnl_addr_delete(socket._sock, self._rtnl_addr, flags)
        if ret < 0:
            raise netlink.KernelError(ret)

    ###################################################################
    # private properties
    #
    # Used for formatting output. USE AT OWN RISK
    @property
    def _flags(self):
        return ",".join(self.flags)

    def format(self, details=False, stats=False, nodev=False, indent=""):
        """Return address as formatted text"""
        fmt = util.MyFormatter(self, indent)

        buf = fmt.format("{a|local!b}")

        if not nodev:
            buf += fmt.format(" {a|ifindex}")

        buf += fmt.format(" {a|scope}")

        if self.label:
            buf += fmt.format(' "{a|label}"')

        buf += fmt.format(" <{a|_flags}>")

        if details:
            buf += fmt.nl("\t{t|broadcast} {t|multicast}") + fmt.nl(
                "\t{t|peer} {t|anycast}"
            )

            if self.valid_lifetime:
                buf += fmt.nl("\t{s|valid-lifetime!k} " "{a|valid_lifetime}")

            if self.preferred_lifetime:
                buf += fmt.nl("\t{s|preferred-lifetime!k} " "{a|preferred_lifetime}")

        if stats and (self.create_time or self.last_update):
            buf += self.nl(
                "\t{s|created!k} {a|create_time}" " {s|last-updated!k} {a|last_update}"
            )

        return buf



#
# Netlink interface based on libnl
#
# Copyright (c) 2011 Thomas Graf <tgraf@suug.ch>
#

"""netlink library based on libnl

This module provides an interface to netlink sockets

The module contains the following public classes:
 - Socket -- The netlink socket
 - Message -- The netlink message
 - Callback -- The netlink callback handler
 - Object -- Abstract object (based on struct nl_obect in libnl) used as
         base class for all object types which can be put into a Cache
 - Cache -- A collection of objects which are derived from the base
        class Object. Used for netlink protocols which maintain a list
        or tree of objects.
 - DumpParams --

The following exceptions are defined:
 - NetlinkError -- Base exception for all general purpose exceptions raised.
 - KernelError -- Raised when the kernel returns an error as response to a
          request.

All other classes or functions in this module are considered implementation
details.
"""
from __future__ import absolute_import


from . import capi
import sys
import socket

__all__ = [
    "Socket",
    "Message",
    "Callback",
    "DumpParams",
    "Object",
    "Cache",
    "KernelError",
    "NetlinkError",
]

__version__ = "0.1"

# netlink protocols
NETLINK_ROUTE = 0
# NETLINK_UNUSED = 1
NETLINK_USERSOCK = 2
NETLINK_FIREWALL = 3
NETLINK_INET_DIAG = 4
NETLINK_NFLOG = 5
NETLINK_XFRM = 6
NETLINK_SELINUX = 7
NETLINK_ISCSI = 8
NETLINK_AUDIT = 9
NETLINK_FIB_LOOKUP = 10
NETLINK_CONNECTOR = 11
NETLINK_NETFILTER = 12
NETLINK_IP6_FW = 13
NETLINK_DNRTMSG = 14
NETLINK_KOBJECT_UEVENT = 15
NETLINK_GENERIC = 16
NETLINK_SCSITRANSPORT = 18
NETLINK_ECRYPTFS = 19

NL_DONTPAD = 0
NL_AUTO_PORT = 0
NL_AUTO_SEQ = 0

NL_DUMP_LINE = 0
NL_DUMP_DETAILS = 1
NL_DUMP_STATS = 2

NLM_F_REQUEST = 1
NLM_F_MULTI = 2
NLM_F_ACK = 4
NLM_F_ECHO = 8

NLM_F_ROOT = 0x100
NLM_F_MATCH = 0x200
NLM_F_ATOMIC = 0x400
NLM_F_DUMP = NLM_F_ROOT | NLM_F_MATCH

NLM_F_REPLACE = 0x100
NLM_F_EXCL = 0x200
NLM_F_CREATE = 0x400
NLM_F_APPEND = 0x800


class NetlinkError(Exception):
    def __init__(self, error):
        self._error = error
        self._msg = capi.nl_geterror(error)

    def __str__(self):
        return self._msg


class KernelError(NetlinkError):
    def __str__(self):
        return "Kernel returned: {0}".format(self._msg)


class ImmutableError(NetlinkError):
    def __init__(self, msg):
        self._msg = msg

    def __str__(self):
        return "Immutable attribute: {0}".format(self._msg)


class Message(object):
    """Netlink message"""

    def __init__(self, size=0):
        if size == 0:
            self._msg = capi.nlmsg_alloc()
        else:
            self._msg = capi.nlmsg_alloc_size(size)

        if self._msg is None:
            raise Exception("Message allocation returned NULL")

    def __del__(self):
        capi.nlmsg_free(self._msg)

    def __len__(self):
        return capi.nlmsg_len(capi.nlmsg_hdr(self._msg))

    @property
    def protocol(self):
        return capi.nlmsg_get_proto(self._msg)

    @protocol.setter
    def protocol(self, value):
        capi.nlmsg_set_proto(self._msg, value)

    @property
    def maxSize(self):
        return capi.nlmsg_get_max_size(self._msg)

    @property
    def hdr(self):
        return capi.nlmsg_hdr(self._msg)

    @property
    def data(self):
        return capi.nlmsg_data(self._msg)

    @property
    def attrs(self):
        return capi.nlmsg_attrdata(self._msg)

    def send(self, sock):
        sock.send(self)


class Callback(object):
    """Netlink callback"""

    def __init__(self, kind=capi.NL_CB_DEFAULT):
        if isinstance(kind, Callback):
            self._cb = capi.py_nl_cb_clone(kind._cb)
        else:
            self._cb = capi.nl_cb_alloc(kind)

    def __del__(self):
        capi.py_nl_cb_put(self._cb)

    def set_type(self, t, k, handler, obj):
        return capi.py_nl_cb_set(self._cb, t, k, handler, obj)

    def set_all(self, k, handler, obj):
        return capi.py_nl_cb_set_all(self._cb, k, handler, obj)

    def set_err(self, k, handler, obj):
        return capi.py_nl_cb_err(self._cb, k, handler, obj)

    def clone(self):
        return Callback(self)


class Socket(object):
    """Netlink socket"""

    def __init__(self, cb=None):
        if isinstance(cb, Callback):
            self._sock = capi.nl_socket_alloc_cb(cb._cb)
        elif cb is None:
            self._sock = capi.nl_socket_alloc()
        else:
            raise Exception("'cb' parameter has wrong type")

        if self._sock is None:
            raise Exception("NULL pointer returned while allocating socket")

    def __del__(self):
        capi.nl_socket_free(self._sock)

    def __str__(self):
        return "nlsock<{0}>".format(self.local_port)

    @property
    def local_port(self):
        return capi.nl_socket_get_local_port(self._sock)

    @local_port.setter
    def local_port(self, value):
        capi.nl_socket_set_local_port(self._sock, int(value))

    @property
    def peer_port(self):
        return capi.nl_socket_get_peer_port(self._sock)

    @peer_port.setter
    def peer_port(self, value):
        capi.nl_socket_set_peer_port(self._sock, int(value))

    @property
    def peer_groups(self):
        return capi.nl_socket_get_peer_groups(self._sock)

    @peer_groups.setter
    def peer_groups(self, value):
        capi.nl_socket_set_peer_groups(self._sock, value)

    def set_bufsize(self, rx, tx):
        capi.nl_socket_set_buffer_size(self._sock, rx, tx)

    def connect(self, proto):
        capi.nl_connect(self._sock, proto)
        return self

    def disconnect(self):
        capi.nl_close(self._sock)

    def sendto(self, buf):
        ret = capi.nl_sendto(self._sock, buf, len(buf))
        if ret < 0:
            raise Exception("Failed to send")
        else:
            return ret

    def send_auto_complete(self, msg):
        if not isinstance(msg, Message):
            raise Exception("must provide Message instance")
        ret = capi.nl_send_auto_complete(self._sock, msg._msg)
        if ret < 0:
            raise Exception("send_auto_complete failed: ret=%d" % ret)
        return ret

    def recvmsgs(self, recv_cb):
        if not isinstance(recv_cb, Callback):
            raise Exception("must provide Callback instance")
        ret = capi.nl_recvmsgs(self._sock, recv_cb._cb)
        if ret < 0:
            raise Exception("recvmsg failed: ret=%d" % ret)


_sockets = {}


def lookup_socket(protocol):
    try:
        sock = _sockets[protocol]
    except KeyError:
        sock = Socket()
        sock.connect(protocol)
        _sockets[protocol] = sock

    return sock


class DumpParams(object):
    """Dumping parameters"""

    def __init__(self, type_=NL_DUMP_LINE):
        self._dp = capi.alloc_dump_params()
        if not self._dp:
            raise Exception("Unable to allocate struct nl_dump_params")

        self._dp.dp_type = type_

    def __del__(self):
        capi.free_dump_params(self._dp)

    @property
    def type(self):
        return self._dp.dp_type

    @type.setter
    def type(self, value):
        self._dp.dp_type = value

    @property
    def prefix(self):
        return self._dp.dp_prefix

    @prefix.setter
    def prefix(self, value):
        self._dp.dp_prefix = value


# underscore this to make sure it is deleted first upon module deletion
_defaultDumpParams = DumpParams(NL_DUMP_LINE)


class Object(object):
    """Cacheable object (base class)"""

    def __init__(self, obj_name, name, obj=None):
        self._obj_name = obj_name
        self._name = name
        self._modules = []

        if not obj:
            obj = capi.object_alloc_name(self._obj_name)

        self._nl_object = obj

        # Create a clone which stores the original state to notice
        # modifications
        clone_obj = capi.nl_object_clone(self._nl_object)
        self._orig = self._obj2type(clone_obj)

    def __del__(self):
        if not self._nl_object:
            raise ValueError()

        capi.nl_object_put(self._nl_object)

    def __str__(self):
        if hasattr(self, "format"):
            return self.format()
        else:
            return capi.nl_object_dump_buf(self._nl_object, 4096).rstrip()

    def _new_instance(self):
        raise NotImplementedError()

    def clone(self):
        """Clone object"""
        return self._new_instance(capi.nl_object_clone(self._nl_object))

    def _module_lookup(self, path, constructor=None):
        """Lookup object specific module and load it

        Object implementations consisting of multiple types may
        offload some type specific code to separate modules which
        are loadable on demand, e.g. a VLAN link or a specific
        queueing discipline implementation.

        Loads the module `path` and calls the constructor if
        supplied or `module`.init()

        The constructor/init function typically assigns a new
        object covering the type specific implementation aspects
        to the new object, e.g. link.vlan = VLANLink()
        """
        try:
            __import__(path)
        except ImportError:
            return

        module = sys.modules[path]

        if constructor:
            ret = getattr(module, constructor)(self)
        else:
            ret = module.init(self)

        if ret:
            self._modules.append(ret)

    def _module_brief(self):
        ret = ""

        for module in self._modules:
            if hasattr(module, "brief"):
                ret += module.brief()

        return ret

    def dump(self, params=None):
        """Dump object as human readable text"""
        if params is None:
            params = _defaultDumpParams

        capi.nl_object_dump(self._nl_object, params._dp)

    @property
    def mark(self):
        return bool(capi.nl_object_is_marked(self._nl_object))

    @mark.setter
    def mark(self, value):
        if value:
            capi.nl_object_mark(self._nl_object)
        else:
            capi.nl_object_unmark(self._nl_object)

    @property
    def shared(self):
        return capi.nl_object_shared(self._nl_object) != 0

    @property
    def attrs(self):
        attr_list = capi.nl_object_attr_list(self._nl_object, 1024)
        return attr_list[0].split()

    @property
    def refcnt(self):
        return capi.nl_object_get_refcnt(self._nl_object)

    # this method resolves multiple levels of sub types to allow
    # accessing properties of subclass/subtypes (e.g. link.vlan.id)
    def _resolve(self, attr):
        obj = self
        lst = attr.split(".")
        while len(lst) > 1:
            obj = getattr(obj, lst.pop(0))
        return (obj, lst.pop(0))

    def _setattr(self, attr, val):
        obj, attr = self._resolve(attr)
        return setattr(obj, attr, val)

    def _hasattr(self, attr):
        obj, attr = self._resolve(attr)
        return hasattr(obj, attr)


class ObjIterator(object):
    def __init__(self, cache, obj):
        self._cache = cache
        self._nl_object = None

        if not obj:
            self._end = 1
        else:
            capi.nl_object_get(obj)
            self._nl_object = obj
            self._first = 1
            self._end = 0

    def __del__(self):
        if self._nl_object:
            capi.nl_object_put(self._nl_object)

    def __iter__(self):
        return self

    def get_next(self):
        return capi.nl_cache_get_next(self._nl_object)

    def next(self):
        return self.__next__()

    def __next__(self):
        if self._end:
            raise StopIteration()

        if self._first:
            ret = self._nl_object
            self._first = 0
        else:
            ret = self.get_next()
            if not ret:
                self._end = 1
                raise StopIteration()

        # return ref of previous element and acquire ref of current
        # element to have object stay around until we fetched the
        # next ptr
        capi.nl_object_put(self._nl_object)
        capi.nl_object_get(ret)
        self._nl_object = ret

        # reference used inside object
        capi.nl_object_get(ret)
        return self._cache._new_object(ret)


class ReverseObjIterator(ObjIterator):
    def get_next(self):
        return capi.nl_cache_get_prev(self._nl_object)


class Cache(object):
    """Collection of netlink objects"""

    def __init__(self):
        if self.__class__ is Cache:
            raise NotImplementedError()
        self.arg1 = None
        self.arg2 = None

    def __del__(self):
        capi.nl_cache_free(self._nl_cache)

    def __len__(self):
        return capi.nl_cache_nitems(self._nl_cache)

    def __iter__(self):
        obj = capi.nl_cache_get_first(self._nl_cache)
        return ObjIterator(self, obj)

    def __reversed__(self):
        obj = capi.nl_cache_get_last(self._nl_cache)
        return ReverseObjIterator(self, obj)

    def __contains__(self, item):
        obj = capi.nl_cache_search(self._nl_cache, item._nl_object)
        if obj is None:
            return False
        else:
            capi.nl_object_put(obj)
            return True

    # called by sub classes to allocate type specific caches by name
    @staticmethod
    def _alloc_cache_name(name):
        return capi.alloc_cache_name(name)

    # implemented by sub classes, must return new instasnce of cacheable
    # object
    @staticmethod
    def _new_object(obj):
        raise NotImplementedError()

    # implemented by sub classes, must return instance of sub class
    def _new_cache(self, cache):
        raise NotImplementedError()

    def subset(self, filter_):
        """Return new cache containing subset of cache

        Cretes a new cache containing all objects which match the
        specified filter.
        """
        if not filter_:
            raise ValueError()

        c = capi.nl_cache_subset(self._nl_cache, filter_._nl_object)
        return self._new_cache(cache=c)

    def dump(self, params=None, filter_=None):
        """Dump (print) cache as human readable text"""
        if not params:
            params = _defaultDumpParams

        if filter_:
            filter_ = filter_._nl_object

        capi.nl_cache_dump_filter(self._nl_cache, params._dp, filter_)

    def clear(self):
        """Remove all cache entries"""
        capi.nl_cache_clear(self._nl_cache)

    # Called by sub classes to set first cache argument
    def _set_arg1(self, arg):
        self.arg1 = arg
        capi.nl_cache_set_arg1(self._nl_cache, arg)

    # Called by sub classes to set second cache argument
    def _set_arg2(self, arg):
        self.arg2 = arg
        capi.nl_cache_set_arg2(self._nl_cache, arg)

    def refill(self, socket=None):
        """Clear cache and refill it"""
        if socket is None:
            socket = lookup_socket(self._protocol)

        capi.nl_cache_refill(socket._sock, self._nl_cache)
        return self

    def resync(self, socket=None, cb=None, args=None):
        """Synchronize cache with content in kernel"""
        if socket is None:
            socket = lookup_socket(self._protocol)

        capi.nl_cache_resync(socket._sock, self._nl_cache, cb, args)

    def provide(self):
        """Provide this cache to others

        Caches which have been "provided" are made available
        to other users (of the same application context) which
        "require" it. F.e. a link cache is generally provided
        to allow others to translate interface indexes to
        link names
        """

        capi.nl_cache_mngt_provide(self._nl_cache)

    def unprovide(self):
        """Unprovide this cache

        No longer make the cache available to others. If the cache
        has been handed out already, that reference will still
        be valid.
        """
        capi.nl_cache_mngt_unprovide(self._nl_cache)


# Cache Manager (Work in Progress)
NL_AUTO_PROVIDE = 1


class CacheManager(object):
    def __init__(self, protocol, flags=None):

        self._sock = Socket()
        self._sock.connect(protocol)

        if not flags:
            flags = NL_AUTO_PROVIDE

        self._mngr = capi.cache_mngr_alloc(self._sock._sock, protocol, flags)

    def __del__(self):
        if self._sock:
            self._sock.disconnect()

        if self._mngr:
            capi.nl_cache_mngr_free(self._mngr)

    def add(self, name):
        capi.cache_mngr_add(self._mngr, name, None, None)


class AddressFamily(object):
    """Address family representation

    af = AddressFamily('inet6')
    # raises:
    #   - ValueError if family name is not known
    #   - TypeError if invalid type is specified for family

    print af        # => 'inet6' (string representation)
    print int(af)   # => 10 (numeric representation)
    print repr(af)  # => AddressFamily('inet6')
    """

    def __init__(self, family=socket.AF_UNSPEC):
        if isinstance(family, str):
            family = capi.nl_str2af(family)
            if family < 0:
                raise ValueError("Unknown family name")
        elif not isinstance(family, int):
            raise TypeError()

        self._family = family

    def __str__(self):
        return capi.nl_af2str(self._family, 32)[0]

    def __int__(self):
        return self._family

    def __repr__(self):
        return "AddressFamily({0!r})".format(str(self))


class AbstractAddress(object):
    """Abstract address object

    addr = AbstractAddress('127.0.0.1/8')
    print addr               # => '127.0.0.1/8'
    print addr.prefixlen     # => '8'
    print addr.family        # => 'inet'
    print len(addr)          # => '4' (32bit ipv4 address)

    a = AbstractAddress('10.0.0.1/24')
    b = AbstractAddress('10.0.0.2/24')
    print a == b             # => False


    """

    def __init__(self, addr):
        self._nl_addr = None

        if isinstance(addr, str):
            # returns None on success I guess
            # TO CORRECT
            addr = capi.addr_parse(addr, socket.AF_UNSPEC)
            if addr is None:
                raise ValueError("Invalid address format")
        elif addr:
            capi.nl_addr_get(addr)

        self._nl_addr = addr

    def __del__(self):
        if self._nl_addr:
            capi.nl_addr_put(self._nl_addr)

    def __cmp__(self, other):
        if isinstance(other, str):
            other = AbstractAddress(other)

        diff = self.prefixlen - other.prefixlen
        if diff == 0:
            diff = capi.nl_addr_cmp(self._nl_addr, other._nl_addr)

        return diff

    def contains(self, item):
        diff = int(self.family) - int(item.family)
        if diff:
            return False

        if item.prefixlen < self.prefixlen:
            return False

        diff = capi.nl_addr_cmp_prefix(self._nl_addr, item._nl_addr)
        return diff == 0

    def __nonzero__(self):
        if self._nl_addr:
            return not capi.nl_addr_iszero(self._nl_addr)
        else:
            return False

    def __len__(self):
        if self._nl_addr:
            return capi.nl_addr_get_len(self._nl_addr)
        else:
            return 0

    def __str__(self):
        if self._nl_addr:
            return capi.nl_addr2str(self._nl_addr, 64)[0]
        else:
            return "none"

    @property
    def shared(self):
        """True if address is shared (multiple users)"""
        if self._nl_addr:
            return capi.nl_addr_shared(self._nl_addr) != 0
        else:
            return False

    @property
    def prefixlen(self):
        """Length of prefix (number of bits)"""
        if self._nl_addr:
            return capi.nl_addr_get_prefixlen(self._nl_addr)
        else:
            return 0

    @prefixlen.setter
    def prefixlen(self, value):
        if not self._nl_addr:
            raise TypeError()

        capi.nl_addr_set_prefixlen(self._nl_addr, int(value))

    @property
    def family(self):
        """Address family"""
        f = 0
        if self._nl_addr:
            f = capi.nl_addr_get_family(self._nl_addr)

        return AddressFamily(f)

    @family.setter
    def family(self, value):
        if not self._nl_addr:
            raise TypeError()

        if not isinstance(value, AddressFamily):
            value = AddressFamily(value)

        capi.nl_addr_set_family(self._nl_addr, int(value))


# keyword:
#   type = { int | str }
#   immutable = { True | False }
#   fmt = func (formatting function)
#   title = string


def nlattr(**kwds):
    """netlink object attribute decorator

    decorator used to mark mutable and immutable properties
    of netlink objects. All properties marked as such are
    regarded to be accessable.

    @property
    @netlink.nlattr(type=int)
    def my_attr(self):
        return self._my_attr

    """

    def wrap_fn(func):
        func.formatinfo = kwds
        return func

    return wrap_fn



#
# Copyright (c) 2011 Thomas Graf <tgraf@suug.ch>
#

"""IPv4

"""

from __future__ import absolute_import

__all__ = []

from ... import core as netlink
from .. import capi as capi
from ... import util as util

DEVCONF_FORWARDING = 1
DEVCONF_MC_FORWARDING = 2
DEVCONF_PROXY_ARP = 3
DEVCONF_ACCEPT_REDIRECTS = 4
DEVCONF_SECURE_REDIRECTS = 5
DEVCONF_SEND_REDIRECTS = 6
DEVCONF_SHARED_MEDIA = 7
DEVCONF_RP_FILTER = 8
DEVCONF_ACCEPT_SOURCE_ROUTE = 9
DEVCONF_BOOTP_RELAY = 10
DEVCONF_LOG_MARTIANS = 11
DEVCONF_TAG = 12
DEVCONF_ARPFILTER = 13
DEVCONF_MEDIUM_ID = 14
DEVCONF_NOXFRM = 15
DEVCONF_NOPOLICY = 16
DEVCONF_FORCE_IGMP_VERSION = 17
DEVCONF_ARP_ANNOUNCE = 18
DEVCONF_ARP_IGNORE = 19
DEVCONF_PROMOTE_SECONDARIES = 20
DEVCONF_ARP_ACCEPT = 21
DEVCONF_ARP_NOTIFY = 22
DEVCONF_ACCEPT_LOCAL = 23
DEVCONF_SRC_VMARK = 24
DEVCONF_PROXY_ARP_PVLAN = 25
DEVCONF_MAX = DEVCONF_PROXY_ARP_PVLAN


def _resolve(id):
    if type(id) is str:
        id = capi.rtnl_link_inet_str2devconf(id)[0]
        if id < 0:
            raise NameError("unknown configuration id")
    return id


class InetLink(object):
    def __init__(self, link):
        self._link = link

    def details(self, fmt):
        buf = fmt.nl("\n\t{0}\n\t".format(util.title("Configuration:")))

        for i in range(DEVCONF_FORWARDING, DEVCONF_MAX + 1):
            if i & 1 and i > 1:
                buf += fmt.nl("\t")
            txt = util.kw(capi.rtnl_link_inet_devconf2str(i, 32)[0])
            buf += fmt.format("{0:28s} {1:12}  ", txt, self.get_conf(i))

        return buf

    def get_conf(self, id):
        return capi.inet_get_conf(self._link._rtnl_link, _resolve(id))

    def set_conf(self, id, value):
        return capi.rtnl_link_inet_set_conf(
            self._link._rtnl_link, _resolve(id), int(value)
        )

    @property
    @netlink.nlattr(type=bool, fmt=util.boolean)
    def forwarding(self):
        return bool(self.get_conf(DEVCONF_FORWARDING))

    @forwarding.setter
    def forwarding(self, value):
        self.set_conf(DEVCONF_FORWARDING, int(value))

    @property
    @netlink.nlattr(type=bool, fmt=util.boolean)
    def mc_forwarding(self):
        return bool(self.get_conf(DEVCONF_MC_FORWARDING))

    @mc_forwarding.setter
    def mc_forwarding(self, value):
        self.set_conf(DEVCONF_MC_FORWARDING, int(value))

    @property
    @netlink.nlattr(type=bool, fmt=util.boolean)
    def proxy_arp(self):
        return bool(self.get_conf(DEVCONF_PROXY_ARP))

    @proxy_arp.setter
    def proxy_arp(self, value):
        self.set_conf(DEVCONF_PROXY_ARP, int(value))

    @property
    @netlink.nlattr(type=bool, fmt=util.boolean)
    def accept_redirects(self):
        return bool(self.get_conf(DEVCONF_ACCEPT_REDIRECTS))

    @accept_redirects.setter
    def accept_redirects(self, value):
        self.set_conf(DEVCONF_ACCEPT_REDIRECTS, int(value))

    @property
    @netlink.nlattr(type=bool, fmt=util.boolean)
    def secure_redirects(self):
        return bool(self.get_conf(DEVCONF_SECURE_REDIRECTS))

    @secure_redirects.setter
    def secure_redirects(self, value):
        self.set_conf(DEVCONF_SECURE_REDIRECTS, int(value))

    @property
    @netlink.nlattr(type=bool, fmt=util.boolean)
    def send_redirects(self):
        return bool(self.get_conf(DEVCONF_SEND_REDIRECTS))

    @send_redirects.setter
    def send_redirects(self, value):
        self.set_conf(DEVCONF_SEND_REDIRECTS, int(value))

    @property
    @netlink.nlattr(type=bool, fmt=util.boolean)
    def shared_media(self):
        return bool(self.get_conf(DEVCONF_SHARED_MEDIA))

    @shared_media.setter
    def shared_media(self, value):
        self.set_conf(DEVCONF_SHARED_MEDIA, int(value))


# 	IPV4_DEVCONF_RP_FILTER,
# 	IPV4_DEVCONF_ACCEPT_SOURCE_ROUTE,
# 	IPV4_DEVCONF_BOOTP_RELAY,
# 	IPV4_DEVCONF_LOG_MARTIANS,
# 	IPV4_DEVCONF_TAG,
# 	IPV4_DEVCONF_ARPFILTER,
# 	IPV4_DEVCONF_MEDIUM_ID,
# 	IPV4_DEVCONF_NOXFRM,
# 	IPV4_DEVCONF_NOPOLICY,
# 	IPV4_DEVCONF_FORCE_IGMP_VERSION,
# 	IPV4_DEVCONF_ARP_ANNOUNCE,
# 	IPV4_DEVCONF_ARP_IGNORE,
# 	IPV4_DEVCONF_PROMOTE_SECONDARIES,
# 	IPV4_DEVCONF_ARP_ACCEPT,
# 	IPV4_DEVCONF_ARP_NOTIFY,
# 	IPV4_DEVCONF_ACCEPT_LOCAL,
# 	IPV4_DEVCONF_SRC_VMARK,
# 	IPV4_DEVCONF_PROXY_ARP_PVLAN,



#
# Copyright (c) 2011 Thomas Graf <tgraf@suug.ch>
#

"""HTB qdisc

"""

from __future__ import absolute_import

from ... import core as netlink
from ... import util as util
from .. import capi as capi
from .. import tc as tc


class HTBQdisc(object):
    def __init__(self, qdisc):
        self._qdisc = qdisc

    @property
    @netlink.nlattr(type=int)
    def default_class(self):
        return tc.Handle(capi.rtnl_htb_get_defcls(self._qdisc._rtnl_qdisc))

    @default_class.setter
    def default_class(self, value):
        capi.rtnl_htb_set_defcls(self._qdisc._rtnl_qdisc, int(value))

    @property
    @netlink.nlattr(type=int)
    def r2q(self):
        return capi.rtnl_htb_get_rate2quantum(self._qdisc._rtnl_qdisc)

    @r2q.setter
    def r2q(self, value):
        capi.rtnl_htb_get_rate2quantum(self._qdisc._rtnl_qdisc, int(value))

    def brief(self):
        fmt = util.MyFormatter(self)

        ret = " {s|default-class!k} {a|default_class}"

        if self.r2q:
            ret += " {s|r2q!k} {a|r2q}"

        return fmt.format(ret)


class HTBClass(object):
    def __init__(self, cl):
        self._class = cl

    @property
    @netlink.nlattr(type=str)
    def rate(self):
        rate = capi.rtnl_htb_get_rate(self._class._rtnl_class)
        return util.Rate(rate)

    @rate.setter
    def rate(self, value):
        capi.rtnl_htb_set_rate(self._class._rtnl_class, int(value))

    @property
    @netlink.nlattr(type=str)
    def ceil(self):
        ceil = capi.rtnl_htb_get_ceil(self._class._rtnl_class)
        return util.Rate(ceil)

    @ceil.setter
    def ceil(self, value):
        capi.rtnl_htb_set_ceil(self._class._rtnl_class, int(value))

    @property
    @netlink.nlattr(type=str)
    def burst(self):
        burst = capi.rtnl_htb_get_rbuffer(self._class._rtnl_class)
        return util.Size(burst)

    @burst.setter
    def burst(self, value):
        capi.rtnl_htb_set_rbuffer(self._class._rtnl_class, int(value))

    @property
    @netlink.nlattr(type=str)
    def ceil_burst(self):
        burst = capi.rtnl_htb_get_cbuffer(self._class._rtnl_class)
        return util.Size(burst)

    @ceil_burst.setter
    def ceil_burst(self, value):
        capi.rtnl_htb_set_cbuffer(self._class._rtnl_class, int(value))

    @property
    @netlink.nlattr(type=int)
    def prio(self):
        return capi.rtnl_htb_get_prio(self._class._rtnl_class)

    @prio.setter
    def prio(self, value):
        capi.rtnl_htb_set_prio(self._class._rtnl_class, int(value))

    @property
    @netlink.nlattr(type=int)
    def quantum(self):
        return capi.rtnl_htb_get_quantum(self._class._rtnl_class)

    @quantum.setter
    def quantum(self, value):
        capi.rtnl_htb_set_quantum(self._class._rtnl_class, int(value))

    @property
    @netlink.nlattr(type=int)
    def level(self):
        return capi.rtnl_htb_get_level(self._class._rtnl_class)

    @level.setter
    def level(self, value):
        capi.rtnl_htb_set_level(self._class._rtnl_class, int(value))

    def brief(self):
        fmt = util.MyFormatter(self)

        ret = " {t|prio} {t|rate}"

        if self.rate != self.ceil:
            ret += " {s|borrow-up-to!k} {a|ceil}"

        ret += " {t|burst}"

        return fmt.format(ret)

    def details(self):
        fmt = util.MyFormatter(self)

        return fmt.nl("\t{t|level} {t|quantum}")


def init_qdisc(qdisc):
    qdisc.htb = HTBQdisc(qdisc)
    return qdisc.htb


def init_class(cl):
    cl.htb = HTBClass(cl)
    return cl.htb


# extern void rtnl_htb_set_quantum(struct rtnl_class *, uint32_t quantum);



#
# Copyright (c) 2011 Thomas Graf <tgraf@suug.ch>
#

"""Dummy

"""
from __future__ import absolute_import

__version__ = "1.0"
__all__ = [
    "init",
]


class DummyLink(object):
    def __init__(self, link):
        self._rtnl_link = link

    @staticmethod
    def brief():
        return "dummy"


def init(link):
    link.dummy = DummyLink(link._rtnl_link)
    return link.dummy



#
# Copyright (c) 2011 Thomas Graf <tgraf@suug.ch>
#
from __future__ import absolute_import

__all__ = [
    "TcCache",
    "Tc",
    "QdiscCache",
    "Qdisc",
    "TcClassCache",
    "TcClass",
]

from .. import core as netlink
from . import capi as capi
from .. import util as util
from . import link as Link

TC_PACKETS = 0
TC_BYTES = 1
TC_RATE_BPS = 2
TC_RATE_PPS = 3
TC_QLEN = 4
TC_BACKLOG = 5
TC_DROPS = 6
TC_REQUEUES = 7
TC_OVERLIMITS = 9

TC_H_ROOT = 0xFFFFFFFF
TC_H_INGRESS = 0xFFFFFFF1

STAT_PACKETS = 0
STAT_BYTES = 1
STAT_RATE_BPS = 2
STAT_RATE_PPS = 3
STAT_QLEN = 4
STAT_BACKLOG = 5
STAT_DROPS = 6
STAT_REQUEUES = 7
STAT_OVERLIMITS = 8
STAT_MAX = STAT_OVERLIMITS


class Handle(object):
    """Traffic control handle

    Representation of a traffic control handle which uniquely identifies
    each traffic control object in its link namespace.

    handle = tc.Handle('10:20')
    handle = tc.handle('root')
    print int(handle)
    print str(handle)
    """

    def __init__(self, val=None):
        if type(val) is str:
            val = capi.tc_str2handle(val)
        elif not val:
            val = 0

        self._val = int(val)

    def __cmp__(self, other):
        if other is None:
            other = 0

        if isinstance(other, Handle):
            return int(self) - int(other)
        elif isinstance(other, int):
            return int(self) - other
        else:
            raise TypeError()

    def __int__(self):
        return self._val

    def __str__(self):
        return capi.rtnl_tc_handle2str(self._val, 64)[0]

    def isroot(self):
        return self._val == TC_H_ROOT or self._val == TC_H_INGRESS


class TcCache(netlink.Cache):
    """Cache of traffic control object"""

    def __getitem__(self, key):
        raise NotImplementedError()


class Tc(netlink.Object):
    def __cmp__(self, other):
        diff = self.ifindex - other.ifindex
        if diff == 0:
            diff = int(self.handle) - int(other.handle)
        return diff

    def _tc_module_lookup(self):
        self._module_lookup(self._module_path + self.kind, "init_" + self._name)

    @property
    def root(self):
        """True if tc object is a root object"""
        return self.parent.isroot()

    @property
    def ifindex(self):
        """interface index"""
        return capi.rtnl_tc_get_ifindex(self._rtnl_tc)

    @ifindex.setter
    def ifindex(self, value):
        capi.rtnl_tc_set_ifindex(self._rtnl_tc, int(value))

    @property
    def link(self):
        link = capi.rtnl_tc_get_link(self._rtnl_tc)
        if not link:
            return None

        return Link.Link.from_capi(link)

    @link.setter
    def link(self, value):
        capi.rtnl_tc_set_link(self._rtnl_tc, value._link)

    @property
    def mtu(self):
        return capi.rtnl_tc_get_mtu(self._rtnl_tc)

    @mtu.setter
    def mtu(self, value):
        capi.rtnl_tc_set_mtu(self._rtnl_tc, int(value))

    @property
    def mpu(self):
        return capi.rtnl_tc_get_mpu(self._rtnl_tc)

    @mpu.setter
    def mpu(self, value):
        capi.rtnl_tc_set_mpu(self._rtnl_tc, int(value))

    @property
    def overhead(self):
        return capi.rtnl_tc_get_overhead(self._rtnl_tc)

    @overhead.setter
    def overhead(self, value):
        capi.rtnl_tc_set_overhead(self._rtnl_tc, int(value))

    @property
    def linktype(self):
        return capi.rtnl_tc_get_linktype(self._rtnl_tc)

    @linktype.setter
    def linktype(self, value):
        capi.rtnl_tc_set_linktype(self._rtnl_tc, int(value))

    @property
    @netlink.nlattr(fmt=util.handle)
    def handle(self):
        return Handle(capi.rtnl_tc_get_handle(self._rtnl_tc))

    @handle.setter
    def handle(self, value):
        capi.rtnl_tc_set_handle(self._rtnl_tc, int(value))

    @property
    @netlink.nlattr(fmt=util.handle)
    def parent(self):
        return Handle(capi.rtnl_tc_get_parent(self._rtnl_tc))

    @parent.setter
    def parent(self, value):
        capi.rtnl_tc_set_parent(self._rtnl_tc, int(value))

    @property
    @netlink.nlattr(fmt=util.bold)
    def kind(self):
        return capi.rtnl_tc_get_kind(self._rtnl_tc)

    @kind.setter
    def kind(self, value):
        capi.rtnl_tc_set_kind(self._rtnl_tc, value)
        self._tc_module_lookup()

    def get_stat(self, id):
        return capi.rtnl_tc_get_stat(self._rtnl_tc, id)

    @property
    def _dev(self):
        buf = util.kw("dev") + " "

        if self.link:
            return buf + util.string(self.link.name)
        else:
            return buf + util.num(self.ifindex)

    def brief(self, title, nodev=False, noparent=False):
        ret = title + " {a|kind} {a|handle}"

        if not nodev:
            ret += " {a|_dev}"

        if not noparent:
            ret += " {t|parent}"

        return ret + self._module_brief()

    @staticmethod
    def details():
        return "{t|mtu} {t|mpu} {t|overhead} {t|linktype}"

    @property
    def packets(self):
        return self.get_stat(STAT_PACKETS)

    @property
    def bytes(self):
        return self.get_stat(STAT_BYTES)

    @property
    def qlen(self):
        return self.get_stat(STAT_QLEN)

    @staticmethod
    def stats(fmt):
        return fmt.nl("{t|packets} {t|bytes} {t|qlen}")


class QdiscCache(netlink.Cache):
    """Cache of qdiscs"""

    def __init__(self, cache=None):
        if not cache:
            cache = self._alloc_cache_name("route/qdisc")

        self._protocol = netlink.NETLINK_ROUTE
        self._nl_cache = cache

    # 	def __getitem__(self, key):
    #        	if type(key) is int:
    #                        link = capi.rtnl_link_get(self._this, key)
    #                elif type(key) is str:
    #                        link = capi.rtnl_link_get_by_name(self._this, key)
    #
    # 		if qdisc is None:
    #                        raise KeyError()
    # 		else:
    #                        return Qdisc._from_capi(capi.qdisc2obj(qdisc))

    @staticmethod
    def _new_object(obj):
        return Qdisc(obj)

    @staticmethod
    def _new_cache(cache):
        return QdiscCache(cache=cache)


class Qdisc(Tc):
    """Queueing discipline"""

    def __init__(self, obj=None):
        netlink.Object.__init__(self, "route/qdisc", "qdisc", obj)
        self._module_path = "netlink.route.qdisc."
        self._rtnl_qdisc = self._obj2type(self._nl_object)
        self._rtnl_tc = capi.obj2tc(self._nl_object)

        if self.kind:
            self._tc_module_lookup()

    @classmethod
    def from_capi(cls, obj):
        return cls(capi.qdisc2obj(obj))

    @staticmethod
    def _obj2type(obj):
        return capi.obj2qdisc(obj)

    @staticmethod
    def _new_instance(obj):
        if not obj:
            raise ValueError()

        return Qdisc(obj)

    @property
    def childs(self):
        ret = []

        if int(self.handle):
            ret += get_cls(self.ifindex, parent=self.handle)

            if self.root:
                ret += get_class(self.ifindex, parent=TC_H_ROOT)

            ret += get_class(self.ifindex, parent=self.handle)

        return ret

    # 	def add(self, socket, flags=None):
    #        	if not flags:
    #                        flags = netlink.NLM_F_CREATE
    #
    # 		ret = capi.rtnl_link_add(socket._sock, self._link, flags)
    # 		if ret < 0:
    # 			raise netlink.KernelError(ret)
    #
    # 	def change(self, socket, flags=0):
    # 		"""Commit changes made to the link object"""
    # 		if not self._orig:
    # 			raise NetlinkError('Original link not available')
    #        	ret = capi.rtnl_link_change(socket._sock, self._orig, self._link, flags)
    #                if ret < 0:
    #                        raise netlink.KernelError(ret)
    #
    # 	def delete(self, socket):
    # 		"""Attempt to delete this link in the kernel"""
    #        	ret = capi.rtnl_link_delete(socket._sock, self._link)
    #                if ret < 0:
    #                        raise netlink.KernelError(ret)

    def format(
        self, details=False, stats=False, nodev=False, noparent=False, indent=""
    ):
        """Return qdisc as formatted text"""
        fmt = util.MyFormatter(self, indent)

        buf = fmt.format(self.brief("qdisc", nodev, noparent))

        if details:
            buf += fmt.nl("\t" + self.details())

        if stats:
            buf += self.stats(fmt)

        # 		if stats:
        # 			l = [['Packets', RX_PACKETS, TX_PACKETS],
        # 			     ['Bytes', RX_BYTES, TX_BYTES],
        # 			     ['Errors', RX_ERRORS, TX_ERRORS],
        # 			     ['Dropped', RX_DROPPED, TX_DROPPED],
        # 			     ['Compressed', RX_COMPRESSED, TX_COMPRESSED],
        # 			     ['FIFO Errors', RX_FIFO_ERR, TX_FIFO_ERR],
        # 			     ['Length Errors', RX_LEN_ERR, None],
        # 			     ['Over Errors', RX_OVER_ERR, None],
        # 			     ['CRC Errors', RX_CRC_ERR, None],
        # 			     ['Frame Errors', RX_FRAME_ERR, None],
        # 			     ['Missed Errors', RX_MISSED_ERR, None],
        # 			     ['Abort Errors', None, TX_ABORT_ERR],
        # 			     ['Carrier Errors', None, TX_CARRIER_ERR],
        # 			     ['Heartbeat Errors', None, TX_HBEAT_ERR],
        # 			     ['Window Errors', None, TX_WIN_ERR],
        # 			     ['Collisions', None, COLLISIONS],
        # 			     ['Multicast', None, MULTICAST],
        # 			     ['', None, None],
        # 			     ['Ipv6:', None, None],
        # 			     ['Packets', IP6_INPKTS, IP6_OUTPKTS],
        # 			     ['Bytes', IP6_INOCTETS, IP6_OUTOCTETS],
        # 			     ['Discards', IP6_INDISCARDS, IP6_OUTDISCARDS],
        # 			     ['Multicast Packets', IP6_INMCASTPKTS, IP6_OUTMCASTPKTS],
        # 			     ['Multicast Bytes', IP6_INMCASTOCTETS, IP6_OUTMCASTOCTETS],
        # 			     ['Broadcast Packets', IP6_INBCASTPKTS, IP6_OUTBCASTPKTS],
        # 			     ['Broadcast Bytes', IP6_INBCASTOCTETS, IP6_OUTBCASTOCTETS],
        # 			     ['Delivers', IP6_INDELIVERS, None],
        # 			     ['Forwarded', None, IP6_OUTFORWDATAGRAMS],
        # 			     ['No Routes', IP6_INNOROUTES, IP6_OUTNOROUTES],
        # 			     ['Header Errors', IP6_INHDRERRORS, None],
        # 			     ['Too Big Errors', IP6_INTOOBIGERRORS, None],
        # 			     ['Address Errors', IP6_INADDRERRORS, None],
        # 			     ['Unknown Protocol', IP6_INUNKNOWNPROTOS, None],
        # 			     ['Truncated Packets', IP6_INTRUNCATEDPKTS, None],
        # 			     ['Reasm Timeouts', IP6_REASMTIMEOUT, None],
        # 			     ['Reasm Requests', IP6_REASMREQDS, None],
        # 			     ['Reasm Failures', IP6_REASMFAILS, None],
        # 			     ['Reasm OK', IP6_REASMOKS, None],
        # 			     ['Frag Created', None, IP6_FRAGCREATES],
        # 			     ['Frag Failures', None, IP6_FRAGFAILS],
        # 			     ['Frag OK', None, IP6_FRAGOKS],
        # 			     ['', None, None],
        # 			     ['ICMPv6:', None, None],
        # 			     ['Messages', ICMP6_INMSGS, ICMP6_OUTMSGS],
        # 			     ['Errors', ICMP6_INERRORS, ICMP6_OUTERRORS]]
        #
        # 			buf += '\n\t%s%s%s%s\n' % (33 * ' ', util.title('RX'),
        #                        			   15 * ' ', util.title('TX'))
        #
        # 			for row in l:
        # 				row[0] = util.kw(row[0])
        #                                row[1] = self.get_stat(row[1]) if row[1] else ''
        #                                row[2] = self.get_stat(row[2]) if row[2] else ''
        # 				buf += '\t{0:27} {1:>16} {2:>16}\n'.format(*row)

        return buf


class TcClassCache(netlink.Cache):
    """Cache of traffic classes"""

    def __init__(self, ifindex, cache=None):
        if not cache:
            cache = self._alloc_cache_name("route/class")

        self._protocol = netlink.NETLINK_ROUTE
        self._nl_cache = cache
        self._set_arg1(ifindex)

    @staticmethod
    def _new_object(obj):
        return TcClass(obj)

    def _new_cache(self, cache):
        return TcClassCache(self.arg1, cache=cache)


class TcClass(Tc):
    """Traffic Class"""

    def __init__(self, obj=None):
        netlink.Object.__init__(self, "route/class", "class", obj)
        self._module_path = "netlink.route.qdisc."
        self._rtnl_class = self._obj2type(self._nl_object)
        self._rtnl_tc = capi.obj2tc(self._nl_object)

        if self.kind:
            self._tc_module_lookup()

    @classmethod
    def from_capi(cls, obj):
        return cls(capi.class2obj(obj))

    @staticmethod
    def _obj2type(obj):
        return capi.obj2class(obj)

    @staticmethod
    def _new_instance(obj):
        if not obj:
            raise ValueError()

        return TcClass(obj)

    @property
    def childs(self):
        ret = []

        # classes can have classifiers, child classes and leaf
        # qdiscs
        ret += get_cls(self.ifindex, parent=self.handle)
        ret += get_class(self.ifindex, parent=self.handle)
        ret += get_qdisc(self.ifindex, parent=self.handle)

        return ret

    def format(
        self, details=False, _stats=False, nodev=False, noparent=False, indent=""
    ):
        """Return class as formatted text"""
        fmt = util.MyFormatter(self, indent)

        buf = fmt.format(self.brief("class", nodev, noparent))

        if details:
            buf += fmt.nl("\t" + self.details())

        return buf


class ClassifierCache(netlink.Cache):
    """Cache of traffic classifiers objects"""

    def __init__(self, ifindex, parent, cache=None):
        if not cache:
            cache = self._alloc_cache_name("route/cls")

        self._protocol = netlink.NETLINK_ROUTE
        self._nl_cache = cache
        self._set_arg1(ifindex)
        self._set_arg2(int(parent))

    @staticmethod
    def _new_object(obj):
        return Classifier(obj)

    def _new_cache(self, cache):
        return ClassifierCache(self.arg1, self.arg2, cache=cache)


class Classifier(Tc):
    """Classifier"""

    def __init__(self, obj=None):
        netlink.Object.__init__(self, "route/cls", "cls", obj)
        self._module_path = "netlink.route.cls."
        self._rtnl_cls = self._obj2type(self._nl_object)
        self._rtnl_tc = capi.obj2tc(self._nl_object)

    @classmethod
    def from_capi(cls, obj):
        return cls(capi.cls2obj(obj))

    @staticmethod
    def _obj2type(obj):
        return capi.obj2cls(obj)

    @staticmethod
    def _new_instance(obj):
        if not obj:
            raise ValueError()

        return Classifier(obj)

    @property
    def priority(self):
        return capi.rtnl_cls_get_prio(self._rtnl_cls)

    @priority.setter
    def priority(self, value):
        capi.rtnl_cls_set_prio(self._rtnl_cls, int(value))

    @property
    def protocol(self):
        return capi.rtnl_cls_get_protocol(self._rtnl_cls)

    @protocol.setter
    def protocol(self, value):
        capi.rtnl_cls_set_protocol(self._rtnl_cls, int(value))

    @property
    def childs(self):
        return []

    def format(
        self, details=False, _stats=False, nodev=False, noparent=False, indent=""
    ):
        """Return class as formatted text"""
        fmt = util.MyFormatter(self, indent)

        buf = fmt.format(self.brief("classifier", nodev, noparent))
        buf += fmt.format(" {t|priority} {t|protocol}")

        if details:
            buf += fmt.nl("\t" + self.details())

        return buf


_qdisc_cache = QdiscCache()


def get_qdisc(ifindex, handle=None, parent=None):
    lst = []

    _qdisc_cache.refill()

    for qdisc in _qdisc_cache:
        if qdisc.ifindex != ifindex:
            continue
        if (handle is not None) and (qdisc.handle != handle):
            continue
        if (parent is not None) and (qdisc.parent != parent):
            continue
        lst.append(qdisc)

    return lst


_class_cache = {}


def get_class(ifindex, parent, handle=None):
    lst = []

    try:
        cache = _class_cache[ifindex]
    except KeyError:
        cache = TcClassCache(ifindex)
        _class_cache[ifindex] = cache

    cache.refill()

    for cl in cache:
        if (parent is not None) and (cl.parent != parent):
            continue
        if (handle is not None) and (cl.handle != handle):
            continue
        lst.append(cl)

    return lst


_cls_cache = {}


def get_cls(ifindex, parent, handle=None):

    chain = _cls_cache.get(ifindex, {})

    try:
        cache = chain[parent]
    except KeyError:
        cache = ClassifierCache(ifindex, parent)
        chain[parent] = cache

    cache.refill()

    if handle is None:
        return list(cache)

    return [cls for cls in cache if cls.handle == handle]
